
#include <ddraw.h>
#include "win32_screen_mgr.h"
#include "win32_GUID.h"


win32_screen_mgr::win32_screen_mgr(HWND window_handle)
{
	wnd_handle = window_handle;
	nes_screen = NULL;
	fullscreen = FALSE;
	emutype = EMUTYPE_NES;

	if(!SetScreenmgr(emutype))
	{
		throw "Error initializing win32 windowed screen manager";
	}
}


win32_screen_mgr::~win32_screen_mgr()
{
	if(nes_screen)
		delete nes_screen;
}


boolean win32_screen_mgr::toggle_fullscreen()
{
	switch(emutype){
	case EMUTYPE_NES:
		return nes_screen->toggle_fullscreen();
		break;
	}
	return FALSE;
}


boolean win32_screen_mgr::is_fullscreen(){
	switch(emutype){
	case EMUTYPE_NES:
		return nes_screen->is_fullscreen();
		break;
	}
	return FALSE;
}


boolean win32_screen_mgr::SetScreenmgr(int emu_type){
	if(!emu_type)
		emu_type = emutype;
	try {
		switch(emu_type){
		case EMUTYPE_NES:
			{
				if(nes_screen){
					nes_screen->SetScreenmgr();
				}
				else{
					win32_NES_screen_mgr* sm;
					sm = new win32_NES_screen_mgr(wnd_handle);
					nes_screen = sm;
				}
			}
			break;
		case EMUTYPE_GB:
			break;
		default:
			throw;
			break;
		}
	} catch(const char* IFDEBUG(s)) {
		LOG(s << endl);
		return FALSE;
	} catch(...) {
		return FALSE;
	}
	emutype = emu_type;
	//  assert_palette();
	return TRUE;
}


uint32 win32_screen_mgr::get_width(){
	switch(emutype){
	case EMUTYPE_NES:
		return nes_screen->get_viewable_width();
		break;
	case EMUTYPE_GB:
		break;
	}
	return 100;
}

uint32 win32_screen_mgr::get_height(){
	switch(emutype){
	case EMUTYPE_NES:
		return nes_screen->get_viewable_height();
		break;
	case EMUTYPE_GB:
		break;
	}
	return 100;
}

void win32_screen_mgr::nnnManuEnable(unsigned char f){
	switch(emutype){
	case EMUTYPE_NES:
		nes_screen->nnnManuEnable(f);
		break;
	case EMUTYPE_GB:
		break;
	}
}


win32_NES_screen_mgr* win32_screen_mgr::Get_Nes_Screenmgr(){
	switch(emutype){
	case EMUTYPE_NES:
		return nes_screen;
		break;
	case EMUTYPE_GB:
		//			delete gb_screen;
		SetScreenmgr(emutype);
		return nes_screen;
		break;
	}
	return nes_screen;
}


/*
void win32_screen_mgr::set_screenmode(unsigned char sm)
{
	if(ScreenMode != sm){
		ScreenMode = sm;
		screen->set_screenmode(sm);
	}
}
*/
